#--
#                 - Mellanox Confidential and Proprietary -
# 
# Copyright (C) Dec 2020, Mellanox Technologies Ltd.  ALL RIGHTS RESERVED.
# 
# Except as specifically permitted herein, no portion of the information,
# including but not limited to object code and source code, may be reproduced,
# modified, distributed, republished or otherwise exploited in any form or by
# any means for any purpose without the prior written permission of Mellanox
# Technologies Ltd. Use of software subject to the terms and conditions
# detailed in the file "LICENSE.txt".
#--

import sys
if sys.version_info.major < 3:
    import commands as sub
else:
    import subprocess as sub
import os
import glob
import re
from mtcr import MstDevice
import math
import regaccess

DEV_TYPE_REG_ADDR = int("0xf0014", 16)
GBOX_DIR = '/dev/mst/gbox'
MTUSB_DEV_PATH = '/dev/mst/'
I2C_DEV_PATH = '/dev/'
GBOX_MNGR_SUFIX = "_gbox_mngr"
I2C_DEV_PREFIX  = "/dev/i2c-"
SYSTEM_BOARD_TYPE = "/sys/devices/virtual/dmi/id/board_name"
GBOX_STR = "_gbox53108"

GB_DEV_ID      = int("0x00000252", 16)
GB_MNGR_DEV_ID = int("0x00000253", 16)
SPECTRUM_3_DEV_ID = int("0x00000250", 16)

GB_MNGR_SLAVE_ADDR = int("0x33", 16)

DEV_SW_PTRN = re.compile(r'.+(mt5\d{4}_pciconf\d)')

QUERY_GEARBOX   = 2
QUERY_LINECARD  = 1
GB_MNGR_INDEX    = 4

def check_if_device_gbox(device_name):
    dev = MstDevice(device_name)

    could_read = True
    try :
        output = dev.read4(DEV_TYPE_REG_ADDR)
    except :
        could_read = False
    dev.close()

    if could_read and output == GB_DEV_ID:
        return True
    return False
    # commnad_str = "mcra {} 0xf0014".format(device_name)


def check_if_device_spectrum3(mst_device):
    could_read = True
    try :
        output = mst_device.read4(DEV_TYPE_REG_ADDR)
    except :
        could_read = False
    mst_device.close()

    if could_read and output == SPECTRUM_3_DEV_ID:
        return True
    return False


def check_if_device_gbox_mngr(device_name):
    dev = MstDevice(device_name)
    could_read = True
    output = None
    try :
        dev.msetI2cSlave(GB_MNGR_SLAVE_ADDR)
        output = dev.read4(DEV_TYPE_REG_ADDR)
    except :
        could_read = False
    dev.close()

    if could_read and output == GB_MNGR_DEV_ID:
        return True
    return False
    # commnad_str = "mcra -s 0x33 {}_gbox 0xf0014".format(device_name)


def create_gbox_dir():
    if not os.path.isdir(GBOX_DIR):
        try:
            os.mkdir(GBOX_DIR)
        except OSError:
            print ("-E- Creation of the directory %s failed" % GBOX_DIR)


def is_gbox_folder_exist():
    return os.path.isdir(GBOX_DIR)


def delete_gbox_empty_dir():
    if os.path.isdir(GBOX_DIR):
        try:
            os.rmdir(GBOX_DIR)
        except OSError:
            print ("-E- Deletion of the directory %s failed" % GBOX_DIR)


def create_file(file_name, parent_device):
    if not os.path.exists(file_name) :
        #os.mknod(file_name)
        parent_file = MTUSB_DEV_PATH + parent_device
        if os.path.islink(parent_file):
            link_name = os.readlink(parent_file)
            os.symlink(link_name, file_name)


def create_files_for_gbox(parent_device, is_mngr):
    if is_mngr:
        file_name = GBOX_DIR + "/" + parent_device  + GBOX_STR + '_ln0_' + 'mngr'
        # check first manager file do not exist
        if not os.path.exists(file_name):
            create_file(file_name, parent_device)
            return 1
    else:
        # check first file with same parent was not created yet
        if len(glob.glob(GBOX_DIR + "/" + parent_device + "*")) == 0:
            for i in range(0,4):
                file_name = GBOX_DIR + "/" + parent_device  + GBOX_STR + '_ln0_' + str(i)
                create_file(file_name, parent_device)
            return 4
    return 0

def get_mtusb_devices():
    gbox_list = []
    gbox_mngr_list = []

    files_list = glob.glob(MTUSB_DEV_PATH + "*")
    for dev in files_list:
        # if mtusb device - is link
        if os.path.islink(dev):
            # check if gbox or mngr
            if check_if_device_gbox(dev) or check_if_device_gbox_mngr(dev):
                dev_name = dev.replace(MTUSB_DEV_PATH, '')
                gbox_list.append(dev_name)
                gbox_mngr_list.append(dev_name)

    return gbox_list, gbox_mngr_list

def create_mtusb_gb_devices():
    gb_devs_created = 0
    gbm_devs_created = 0

    gbox_list, gbox_mngr_list = get_mtusb_devices()

    if (len(gbox_list) + len(gbox_mngr_list)) > 0:
        create_gbox_dir()
        # create gbox 0-3 for each mtusb
        for dev in gbox_list:
            gb_devs_created += create_files_for_gbox(dev, False)
        # # create mngr for mtusb
        for dev in gbox_mngr_list:
            gbm_devs_created += create_files_for_gbox(dev, True)

    return gb_devs_created, gbm_devs_created


def is_bufallo():
    cmd = "cat " + SYSTEM_BOARD_TYPE

    rc, output = sub.getstatusoutput(cmd)
    if not rc:
        return (output == "VMOD0011")
    return False


def create_i2c_gb_devices():
    #if not buffalo - return 0 devices
    if not is_bufallo():
        return 0, 0

    created_devices = 0
    created_mngr_devices = 0
    files = glob.glob(I2C_DEV_PATH + "*")
    i2c_devices = []
    for file_name in files:
        if I2C_DEV_PREFIX in file_name:
            i2c_name = file_name.replace(I2C_DEV_PREFIX, '')
            if int(i2c_name) > 100:
                i2c_devices.append(i2c_name)

    gbox_already_exist = is_gbox_folder_exist()

    # create /dev/mst/gbox folder if not exist yet
    if len(i2c_devices) > 0 and not gbox_already_exist:
        create_gbox_dir()

    for i2c_num in i2c_devices:
        add_mngr = 0
        add_gb = 0
        lc_index = math.floor(int(i2c_num) / 100)
        gb_index = int(i2c_num) - (lc_index * 100) - 32 

        # check gb index and line card index
        # if AGBM so 48-32 or 16-32
        file_name = ""
        if (lc_index > 0) and (lc_index < 9):         
            if (((gb_index >= 0) and (gb_index < 4))):
                file_name = GBOX_DIR + "/" + "i2c" + GBOX_STR + "_ln" + str(lc_index) + "_" + str(gb_index)
                add_gb = 1
            if ((gb_index == 16) or (gb_index == -16)):
                file_name = GBOX_DIR + "/" + "i2c" + GBOX_STR + "_ln" + str(lc_index) + "_mngr"
                add_mngr = 1

            # if file do not exist - create it has link
            if file_name and not os.path.exists(file_name) :
                link_name = I2C_DEV_PREFIX + i2c_num
                os.symlink(link_name, file_name)
                created_devices += add_gb
                created_mngr_devices += add_mngr

    # delete folder if none devices created
    if not created_devices and not gbox_already_exist:
        delete_gbox_empty_dir()

    return created_devices, created_mngr_devices


def get_devices_on_slot(reg_access, slot_index):
    devices_on_slot = []
    request_message_sequence = 0
    while True:
        mddq_get = reg_access.sendMddq(QUERY_GEARBOX, request_message_sequence, slot_index)
        last_msg = mddq_get[0]["lst"]
        if mddq_get[0]["lst"] or mddq_get[0]["mid"] or last_msg:
            if mddq_get[1]["device_type"] == 0: # 0 for GB
                devices_on_slot.append(mddq_get[1]["device_index"])
        if last_msg or (not mddq_get[0]["lst"] and not mddq_get[0]["mid"] and not last_msg):
            break
        request_message_sequence += 1
    return devices_on_slot


def query_switch_gbs(mst_str):
    reg_access = regaccess.RegAccess(pci_device=mst_str)
    # query_type: (1: slot information, 2: devices on slot))
    # request_message_sequence: (slots or devices indicator, if set, we have a device)
    # slot_index: line card index
    # query_index: not related to GBs
    slot_index=1
    slots={}
    switch_code = DEV_SW_PTRN.match(mst_str).group(1)
    # query slot information
    try:
        while True:
            mddq_get = reg_access.sendMddq(QUERY_LINECARD, 0, slot_index)
            last_msg = mddq_get[0]["lst"]
            if mddq_get[0]["lst"] or mddq_get[0]["mid"] or last_msg:
                 if mddq_get[1]["ready"]:
                    slots.update({slot_index:get_devices_on_slot(reg_access, slot_index)})
            if last_msg or (not mddq_get[0]["lst"] and not mddq_get[0]["mid"] and not mddq_get[0]["lst"]):
                break
            slot_index += 1
    except regaccess.RegAccException:
        pass
    num_of_gboxs = 0
    num_of_mgboxs = 0
    for lc in slots.keys():
        for gbox in slots[lc]:
            sfx = str(gbox)
            if gbox == GB_MNGR_INDEX:
                sfx = "mngr"
                num_of_mgboxs += 1
            else:
                num_of_gboxs += 1
            file_name = GBOX_DIR + "/" + "switch_" + switch_code + GBOX_STR + "_ln" + str(lc) + "_" + sfx
            # if file do not exist - create it as link
            if file_name and not os.path.exists(file_name) :
                os.symlink(mst_str, file_name)
    return num_of_gboxs, num_of_mgboxs

def create_switch_gb_devices():
    switches_list = []
    for dev in glob.glob("/dev/mst/*"):
        if DEV_SW_PTRN.match(dev):
            switches_list.append(dev)

    create_gbox_dir()

    num_of_gboxs = 0
    num_of_mgboxs = 0
    for switch in switches_list:
        mst_dev = MstDevice(switch)
        if not check_if_device_spectrum3(mst_dev):
            mst_dev.close()
            continue
        mst_dev.close()
        rc = query_switch_gbs(switch)
        num_of_gboxs += rc[0]
        num_of_mgboxs += rc[1]

    if num_of_gboxs == 0:
        delete_gbox_empty_dir()
    
    return num_of_gboxs, num_of_mgboxs


if __name__ == '__main__':

    gb_mtusb_devs_count, gbm_mtusb_devs_count = create_mtusb_gb_devices()

    gb_i2c_devs_count, gbm_i2c_devs_count = create_i2c_gb_devices()

    gb_switch_devs_count, gbm_switch_devs_count = create_switch_gb_devices()

    gb_devs_count = gb_mtusb_devs_count + gb_i2c_devs_count + gb_switch_devs_count
    gbm_devs_count = gbm_mtusb_devs_count + gbm_i2c_devs_count + gbm_switch_devs_count

    if gb_devs_count > 0:
        print ("-I- {0} Gearbox devices added".format(gb_devs_count))
    else: 
        print ("-I- No new Gearbox devices found")
    if gbm_devs_count > 0:
        print ("-I- {0} Gearbox Mannager devices added".format(gbm_devs_count))
